import { onEvent } from "../indexHelpers";
import WUPPopupElement from "../popup/popupElement";
import WUPBaseControl from "./baseControl";
import WUPTextControl from "./text";
export var ShowCases;
(function (ShowCases) {
    /** When $showMenu() called programmatically; Don't use it for $options (it's for nested cycle) */
    ShowCases[ShowCases["onManualCall"] = 1] = "onManualCall";
    /** When control got focus; ignores case when user changed value by click on item on clearButton
     * to change such behavior update WUPBaseComboControl.prototype.canShowMenu */
    ShowCases[ShowCases["onFocus"] = 2] = "onFocus";
    /** When user clicks on control (beside editable not-empty input) */
    ShowCases[ShowCases["onClick"] = 4] = "onClick";
    /** When user clicks on input (by default it's disabled rule to allow user to work with input without popup hide/show blinks) */
    ShowCases[ShowCases["onClickInput"] = 16] = "onClickInput";
    /** When user types text */
    ShowCases[ShowCases["onInput"] = 32] = "onInput";
    /** When user presses arrowUp/arrowDown */
    ShowCases[ShowCases["onPressArrowKey"] = 64] = "onPressArrowKey";
})(ShowCases || (ShowCases = {}));
export var HideCases;
(function (HideCases) {
    HideCases[HideCases["onManualCall"] = 0] = "onManualCall";
    HideCases[HideCases["onClick"] = 1] = "onClick";
    HideCases[HideCases["onSelect"] = 2] = "onSelect";
    HideCases[HideCases["onFocusLost"] = 3] = "onFocusLost";
    HideCases[HideCases["OnPressEsc"] = 4] = "OnPressEsc";
    HideCases[HideCases["OnPressEnter"] = 5] = "OnPressEnter";
    HideCases[HideCases["OnOptionsChange"] = 6] = "OnOptionsChange";
})(HideCases || (HideCases = {}));
/** Base abstract form-control for any control with popup-menu (Dropdown, Datepicker etc.) */
export default class WUPBaseComboControl extends WUPTextControl {
    /** Returns this.constructor // watch-fix: https://github.com/Microsoft/TypeScript/issues/3841#issuecomment-337560146 */
    #ctr = this.constructor;
    static get observedOptions() {
        const arr = super.observedOptions;
        arr.push("readOnlyInput");
        return arr;
    }
    static get $style() {
        return `${super.$style}
      :host {
        cursor: pointer;
      }
      :host input {
        cursor: text;
      }
      :host input:placeholder-shown,
      :host input:read-only {
        cursor: pointer;
      }
      :host label::after {
        content: "";
        -webkit-mask-image: var(--ctrl-icon-img);
        mask-image: var(--ctrl-icon-img);
      }
      @media not all and (prefers-reduced-motion) {
        :host label::after {
          transition: transform var(--anim);
        }
      }
      :host button[clear] {
        margin: 0;
      }
      :host > [menu] {
        padding: 0;
        max-height: 300px;
        z-index: 90010;
      }`;
    }
    static $defaults = {
        ...WUPTextControl.$defaults,
        validationRules: {
            ...WUPBaseControl.$defaults.validationRules,
            _mask: WUPTextControl.$defaults.validationRules._mask,
            _parse: WUPTextControl.$defaults.validationRules._parse,
        },
        showCase: 4 /* ShowCases.onClick */ | 2 /* ShowCases.onFocus */ | 64 /* ShowCases.onPressArrowKey */,
    };
    /** Fires after popup-menu is shown (after animation finishes) */
    $onShowMenu;
    /** Fires after popup is hidden (after animation finishes) */
    $onHideMenu;
    #isShown = false;
    /** Returns if popup-menu is opened */
    get $isShown() {
        return this.#isShown;
    }
    /** Hide popup-menu
     * @returns Promise resolved by animation time */
    async $hideMenu() {
        await this.goHideMenu(0 /* HideCases.onManualCall */);
    }
    /** Show popup-menu
     * @returns Promise resolved resolved by animation time */
    async $showMenu() {
        await this.goShowMenu(1 /* ShowCases.onManualCall */);
        this.#isShown && (await this.$refPopup.$show()); // wait for popup show-end
    }
    /** Reference to popupMenu */
    $refPopup;
    renderControl() {
        super.renderControl();
        const i = this.$refInput;
        i.setAttribute("role", "combobox");
        i.setAttribute("aria-haspopup", "listbox");
        i.setAttribute("aria-expanded", false);
    }
    gotChanges(propsChanged) {
        super.gotChanges(propsChanged);
        this._opts.readOnlyInput
            ? this.$refInput.removeAttribute("aria-autocomplete")
            : this.$refInput.setAttribute("aria-autocomplete", "list");
    }
    gotFormChanges(propsChanged) {
        super.gotFormChanges(propsChanged);
        const isMenuEnabled = !this.$isDisabled && !this.$isReadOnly;
        !isMenuEnabled && this.removePopup();
    }
    setupInputReadonly() {
        this.$refInput.readOnly = this.$isReadOnly || !!this._opts.readOnlyInput;
    }
    /** Returns whether possible to use input-click as dropdown behavior */
    isLockInputDrodpown(e) {
        return (!(this._opts.showCase & 16 /* ShowCases.onClickInput */) &&
            e.type === "click" &&
            e.target instanceof HTMLInputElement &&
            !e.target.readOnly &&
            e.target.value !== "");
    }
    /** Override to change show-behavior */
    canShowMenu(showCase, e) {
        if (this.$isReadOnly || this.$isDisabled) {
            return false;
        }
        if (e && this.isLockInputDrodpown(e)) {
            return false; // if input readonly > dropdown behavior otherwise allow to work with input instead of opening window
        }
        if (showCase === 4 /* ShowCases.onClick */) {
            if (!(this._opts.showCase & 4 /* ShowCases.onClick */)) {
                return false;
            }
        }
        else if (showCase === 2 /* ShowCases.onFocus */) {
            if (!(this._opts.showCase & 2 /* ShowCases.onFocus */)) {
                return false;
            }
            const was = this.$value;
            return new Promise((res) => {
                // WARN: if click on btnClear with control.emptyValue - no changes and popup appeares
                setTimeout(() => {
                    const isChangedByClick = !this.#ctr.$isEqual(was, this.$value);
                    return res(!isChangedByClick); // click on item or btnClear & control gets focus: need to not open menu because user makes action for changing value
                });
            });
        }
        return true;
    }
    async goShowMenu(showCase, e) {
        if (this.#isShown) {
            return this.$refPopup;
        }
        let can = this.canShowMenu(showCase, e);
        if (can instanceof Promise) {
            can = await can;
        }
        if (!can) {
            return null;
        }
        this.#isShown = true;
        // this.$hideError(); // it resolves overflow menu vs error
        if (!this.$refPopup) {
            const p = document.createElement("wup-popup");
            this.$refPopup = p;
            p.$options.showCase = 8 /* PopupShowCases.alwaysOff */;
            p.$options.target = this;
            p.$options.offsetFitElement = [1, 1];
            p.$options.minWidthByTarget = true;
            p.$options.placement = [
                WUPPopupElement.$placements.$bottom.$start,
                WUPPopupElement.$placements.$bottom.$end,
                WUPPopupElement.$placements.$top.$start,
                WUPPopupElement.$placements.$top.$end,
                WUPPopupElement.$placements.$bottom.$start.$resizeHeight,
                WUPPopupElement.$placements.$bottom.$end.$resizeHeight,
                WUPPopupElement.$placements.$top.$start.$resizeHeight,
                WUPPopupElement.$placements.$top.$end.$resizeHeight,
            ];
            p.setAttribute("menu", "");
            p.$options.animation = 1 /* Animations.drawer */;
            const menuId = this.#ctr.$uniqueId;
            const i = this.$refInput;
            i.setAttribute("aria-owns", menuId);
            i.setAttribute("aria-controls", menuId);
            // const wasFcs = this.$isFocused;
            this.renderMenu(p, menuId);
            // rollback the logic if renderMenu() returns Promise
            // const fcs = this.$isFocused;
            // if (!fcs && wasFcs) {
            //   this.#isShown = false;
            //   this.removePopup(); // fix case when user waited for loading and moved focus to another
            //   return null;
            // }
            this.appendChild(p); // WARN: it will show onInit
            // eslint-disable-next-line no-promise-executor-return
            await new Promise((res) => setTimeout(res)); // wait for appending to body so size is defined and possible to scroll
        }
        if (!this.#isShown) {
            return null; // possible when user calls show & hide sync
        }
        this.$refPopup.$show().then(() => this.fireEvent("$showMenu", { cancelable: false }));
        this.setAttribute("opened", ""); // possible when user calls show & hide sync
        this.$refInput.setAttribute("aria-expanded", true);
        // await r; // WARN: it's important don't wait for animation to assign onShow events fast
        return this.$refPopup;
    }
    /** Override to change hide-behavior */
    canHideMenu(hideCase, e) {
        if (e && this.isLockInputDrodpown(e)) {
            return false; // if input readonly > dropdown behavior otherwise allow to work with input instead of opening window
        }
        if (hideCase === 1 /* HideCases.onClick */) {
            if (!(this._opts.showCase & 4 /* ShowCases.onClick */)) {
                return false;
            }
            if (this.$refPopup.includesTarget(e)) {
                return false; // popupClick
            }
        }
        // always hide by focusLost
        // if (hideCase === HideCases.onFocusLost && !(this._opts.showCase & ShowCases.onFocus)) {
        //   return false;
        // }
        return true;
    }
    _isHiding;
    async goHideMenu(hideCase, e) {
        if (!this.#isShown || this._isHiding) {
            return false;
        }
        if (!this.canHideMenu(hideCase, e)) {
            return false;
        }
        this.#isShown = false;
        this._isHiding = true;
        await this.$refPopup?.$hide();
        delete this._isHiding;
        if (this.#isShown) {
            return false; // possible when popup opened again during the animation
        }
        // remove popup only by focusOut to optimize resources
        if (hideCase === 3 /* HideCases.onFocusLost */) {
            this.removePopup();
        }
        this.removeAttribute("opened");
        this.$refInput.setAttribute("aria-expanded", false);
        this.focusMenuItem(null);
        this.selectMenuItem(null);
        setTimeout(() => this.fireEvent("$hideMenu", { cancelable: false }));
        return true;
    }
    /** Current focused menu-item (via aria-activedescendant) */
    _focusedMenuItem;
    /** Focus/resetFocus for item (via aria-activedescendant) */
    focusMenuItem(next) {
        this._focusedMenuItem?.removeAttribute("focused");
        if (next) {
            next.id = next.id || this.#ctr.$uniqueId;
            next.setAttribute("focused", "");
            this.$refInput.setAttribute("aria-activedescendant", next.id);
            this.tryScrollTo(next);
        }
        else {
            this.$refInput.removeAttribute("aria-activedescendant");
        }
        this._focusedMenuItem = next;
    }
    #scrolltid;
    #scrollFrame;
    /** Scroll to element if previous scroll is not in processing (debounce by empty timeout) */
    tryScrollTo(el) {
        if (this.#scrolltid) {
            return;
        }
        this.#scrolltid = setTimeout(() => {
            this.#scrolltid = undefined;
            this.#scrollFrame && window.cancelAnimationFrame(this.#scrollFrame); // cancel previous scrolling
            this.#scrollFrame = undefined;
            const act = () => {
                const p = el.parentElement; // WARN: expected that parent is scrollable
                const prev = p.scrollTop;
                const ifneed = el.scrollIntoViewIfNeeded;
                ifneed ? ifneed.call(el, false) : el.scrollIntoView({ behavior: "instant" });
                if (p.scrollTop !== prev) {
                    this.#scrollFrame = window.requestAnimationFrame(act); // try scroll again later because during the popup animation scrolling can be wrong
                }
                else {
                    this.#scrollFrame = undefined;
                }
            };
            act();
        });
    }
    _selectedMenuItem;
    /** Select item (set aria-selected and scroll to) */
    selectMenuItem(next) {
        this._selectedMenuItem?.setAttribute("aria-selected", false);
        if (next) {
            next.setAttribute("aria-selected", true);
            this.tryScrollTo(next);
        }
        this._selectedMenuItem = next;
    }
    gotKeyDown(e) {
        // don't allow to process Esc-key when menu is opened
        const isEscPrevent = this.#isShown && e.key === "Escape";
        !isEscPrevent && super.gotKeyDown(e);
        if (e.altKey || e.shiftKey || e.ctrlKey) {
            return;
        }
        if (!this.#isShown) {
            if (this._opts.showCase & 64 /* ShowCases.onPressArrowKey */) {
                if (e.key === "ArrowDown" || e.key === "ArrowUp") {
                    e.preventDefault(); // to prevent parent-scroll
                    this.goShowMenu(64 /* ShowCases.onPressArrowKey */, e).then(() => this.focusMenuItemByKeydown(e));
                    return;
                }
            }
            return;
        }
        switch (e.key) {
            case "Escape":
                e.preventDefault();
                this.resetInputValue();
                this.goHideMenu(4 /* HideCases.OnPressEsc */);
                break;
            case "Enter":
                e.preventDefault();
                {
                    const el = this._focusedMenuItem;
                    setTimeout(() => {
                        const isPrevented = el &&
                            // !el.hasAttribute("disabled") && // looks impossible case
                            !el.dispatchEvent(new MouseEvent("click", { cancelable: true, bubbles: true }));
                        if (!isPrevented) {
                            this.$refInput.value || this.$isRequired
                                ? this.resetInputValue()
                                : this.setValue(undefined, 3 /* SetValueReasons.userInput */, true); // reset only when input not empty otherwise user cleared this manually
                            this.goHideMenu(5 /* HideCases.OnPressEnter */);
                        }
                    });
                }
                break;
            case " ":
                break; // don't process " " because it can fire clickEvent on buttons
            default:
                this.focusMenuItemByKeydown(e);
                break;
        }
    }
    gotInput(e, preventValueChange) {
        // gotInput possible on browser-autofill so we need filter check if isFocused
        !this.$isShown && this._opts.showCase & 32 /* ShowCases.onInput */ && this.$isFocused && this.goShowMenu(32 /* ShowCases.onInput */);
        if (preventValueChange) {
            this._refHistory?.handleInput(e);
        }
        else {
            super.gotInput(e);
        }
    }
    gotFocus(ev) {
        const arr = super.gotFocus(ev);
        const r = this.goShowMenu(2 /* ShowCases.onFocus */, ev);
        let clickAfterFocus = true; // prevent clickAfterFocus
        let lblClick = false; // fix when labelOnClick > inputOnClick > inputOnFocus
        let onInputStartClick = false; // mouseDown>selectText>mouseUp - without filter it closes/opens popup when user tries to select text
        const dsps0 = onEvent(this.$refInput, "mousedown", () => {
            const allow = !this.$refInput.readOnly && !(this._opts.showCase & 16 /* ShowCases.onClickInput */);
            onInputStartClick = allow;
        });
        const dsps1 = onEvent(this, "click", (e) => {
            const skip = e.defaultPrevented || e.button || lblClick || onInputStartClick; // e.button > 0 if not left-click
            if (!skip) {
                if (clickAfterFocus) {
                    r.finally(() => this.$isFocused && this.goShowMenu(4 /* ShowCases.onClick */, e)); // menu must be opened if openByFocus is rejected
                }
                else {
                    !this.#isShown ? this.goShowMenu(4 /* ShowCases.onClick */, e) : this.goHideMenu(1 /* HideCases.onClick */, e);
                }
            }
            lblClick = setTimeout(() => (lblClick = false));
        });
        const dsps2 = onEvent(document, "click", (e) => {
            !lblClick && !onInputStartClick && setTimeout(() => this.$isFocused && this.goHideMenu(1 /* HideCases.onClick */, e)); // timeout to prevent show before focusLost
            onInputStartClick = false;
        });
        document.addEventListener("pointerdown", () => {
            clickAfterFocus = false;
        }, { once: true, passive: true, capture: true });
        arr.push(dsps0, dsps1, dsps2);
        return arr;
    }
    gotFocusLost() {
        !this.#isShown && !this._isHiding && this.removePopup(); // otherwise it's removed by hidingMenu
        this.goHideMenu(3 /* HideCases.onFocusLost */);
        this.resetInputValue(); // to update/rollback input according to result
        super.gotFocusLost();
    }
    /** Called when user selected new value from menu and need to hide menu */
    selectValue(v, canHideMenu = true) {
        this.setValue(v, 4 /* SetValueReasons.userSelect */);
        canHideMenu && setTimeout(() => this.goHideMenu(2 /* HideCases.onSelect */)); // without timeout it handles click by listener and opens again
    }
    /** Reset input to currentValue; called on focusOut, pressing Escape or Enter */
    resetInputValue() {
        const isClear = !this.$refInput.value && !this._selectedMenuItem; // if user cleared input and $hided menu - need to clearValue
        isClear
            ? this.setValue(undefined, 2 /* SetValueReasons.clear */)
            : this.setInputValue(this.$value, 4 /* SetValueReasons.userSelect */);
    }
    // @ts-expect-error - because expected string
    setInputValue(v, reason) {
        const p = this.valueToInput(v);
        this.setInputValueDirect(p, reason);
    }
    setInputValueDirect(v, reason) {
        super.setInputValue(v, reason);
    }
    /** Called when popup must be removed (by focus out OR if control removed) */
    removePopup() {
        this.$refPopup?.remove();
        this.$refPopup = undefined;
        this._focusedMenuItem = undefined;
        this._selectedMenuItem = undefined;
        if (this.#isShown) {
            this.#isShown = false;
            this.removeAttribute("opened");
            this.$refInput.setAttribute("aria-expanded", false);
            this.$refInput.removeAttribute("aria-activedescendant");
        }
    }
    gotRemoved() {
        this.removePopup();
        super.gotRemoved();
    }
}
// WARN about chaining
/*
 $showMenu().then(()=> console.warn('done'))
 isShown: true >>> create popup,
 isShowing: true
 ...fetch menu-items (takes time)... >>> render popup with items
 ...animation (takes time)...
 isShowing: false
 >>> console.warn('done')
 show-event
 */
/*
 $hideMenu().then(()=> console.warn('done'))
 isShown: false
 isHiding: true
 ...animation...
 isHiding: false >>> remove popup if focus out
 >>> console.warn('done')
 hide-event
 */
// NiceToHave: option for press-Escape: hideMenu + rollback value to that was before showing OR only hideMenu; now WUPTime.$options.menuButtons changes such behavior
