import WUPBaseControl, { SetValueReasons } from "./baseControl";
declare const tagName = "wup-switch";
declare global {
    namespace WUP.Switch {
        interface EventMap extends WUP.BaseControl.EventMap {
        }
        interface ValidityMap extends WUP.BaseControl.ValidityMap {
        }
        interface Defaults<T = boolean, VM = ValidityMap> extends WUP.BaseControl.Defaults<T, VM> {
        }
        interface Options<T = boolean, VM = ValidityMap> extends WUP.BaseControl.Options<T, VM>, Defaults<T, VM> {
            /** Reversed-style (switch+label for true vs label+switch)
             * @defaultValue false */
            reverse?: boolean;
        }
        interface Attributes extends WUP.BaseControl.Attributes {
            /** InitValue for control */
            defaultChecked?: boolean;
            /** Reversed-style (switch+label vs label+switch) */
            reverse?: boolean | "";
        }
        interface JSXProps<C = WUPSwitchControl> extends WUP.BaseControl.JSXProps<C>, Attributes {
        }
    }
    interface HTMLElementTagNameMap {
        [tagName]: WUPSwitchControl;
    }
    namespace JSX {
        interface IntrinsicElements {
            /** Form-control with toggle button
             *  @see {@link WUPSwitchControl} */
            [tagName]: WUP.Switch.JSXProps;
        }
    }
}
/** Form-control with toggle button
 * @example
  const el = document.createElement("wup-switch");
  el.$options.name = "isDarkMode";
  el.$options.label = "Dark Mode";
  el.$initValue = false;
  const form = document.body.appendChild(document.createElement("wup-form"));
  form.appendChild(el);
  // or HTML
  <wup-form>
    <wup-switch name="isDarkMode" label="Dark Mode" initvalue="false"/>
  </wup-form>;
 * @tutorial innerHTML @example
 * <label>
 *   <span> // extra span requires to use with icons via label:before, label:after without adjustments
 *      <input/>
 *      <strong>{$options.label}</strong>
 *      <span></span> // this icon
 *   </span>
 * </label>
 */
export default class WUPSwitchControl<TOptions extends WUP.Switch.Options = WUP.Switch.Options, EventMap extends WUP.Switch.EventMap = WUP.Switch.EventMap> extends WUPBaseControl<boolean, TOptions, EventMap> {
    #private;
    static get $styleRoot(): string;
    static get $style(): string;
    static get observedOptions(): Array<string>;
    static get observedAttributes(): Array<string>;
    static $isEqual(v1: boolean | undefined, v2: boolean | undefined): boolean;
    static $defaults: WUP.Switch.Defaults;
    get $value(): boolean;
    set $value(v: boolean);
    parse(text: string): boolean | undefined;
    valueToUrl(v: boolean): string | null;
    protected renderControl(): void;
    protected gotReady(): void;
    /** Called when user changes value via click or keyboard */
    protected gotInput(e: Event): void;
    protected setValue(v: boolean, reason: SetValueReasons): boolean | null;
    protected gotChanges(propsChanged: Array<keyof WUP.Switch.Options | any> | null): void;
    protected gotOptionsChanged(e: WUP.Base.OptionEvent): void;
    gotFormChanges(propsChanged: Array<keyof WUP.Form.Options> | null): void;
}
export {};
