// eslint-disable-next-line max-classes-per-file
import { WUPcssHidden } from "../styles";
import WUPBaseControl from "./baseControl";
const tagName = "wup-switch";
/** Form-control with toggle button
 * @example
  const el = document.createElement("wup-switch");
  el.$options.name = "isDarkMode";
  el.$options.label = "Dark Mode";
  el.$initValue = false;
  const form = document.body.appendChild(document.createElement("wup-form"));
  form.appendChild(el);
  // or HTML
  <wup-form>
    <wup-switch name="isDarkMode" label="Dark Mode" initvalue="false"/>
  </wup-form>;
 * @tutorial innerHTML @example
 * <label>
 *   <span> // extra span requires to use with icons via label:before, label:after without adjustments
 *      <input/>
 *      <strong>{$options.label}</strong>
 *      <span></span> // this icon
 *   </span>
 * </label>
 */
export default class WUPSwitchControl extends WUPBaseControl {
    #ctr = this.constructor;
    static get $styleRoot() {
        return `:root {
      --ctrl-switch-padding: 1em;
      --ctrl-switch-on: #fff;
      --ctrl-switch-off: #fff;
      --ctrl-switch-shadow: #0003;
      --ctrl-switch-off-bg: #9f9f9f;
      --ctrl-switch-on-bg: var(--ctrl-focus);
      --ctrl-switch-size-h: var(--ctrl-icon-size);
      --ctrl-switch-size-w: calc(var(--ctrl-switch-size-spot) * 2);
      --ctrl-switch-size-spot: calc(var(--ctrl-switch-size-h) * 1.4);
     }`;
    }
    static get $style() {
        return `${super.$style}
      :host {
        background: none;
        cursor: pointer;
      }
      :host[readonly] {
        cursor: initial;
      }
      :host label {
        padding: var(--ctrl-switch-padding);
      }
      :host strong {
        box-sizing: border-box;
        text-overflow: ellipsis;
        overflow: hidden;
        white-space: nowrap;
        font-weight: normal;
        text-decoration: none;
        color: var(--ctrl-label);
      }
      :host label>span {
        margin-left: 0.5em;
        position: relative;
        width: var(--ctrl-switch-size-w);
        min-width: var(--ctrl-switch-size-w);
        height: var(--ctrl-switch-size-h);
        border-radius: 999px;
        background: var(--ctrl-switch-off-bg);
      }
      :host label>span:before {
        content: "";
        position: absolute;
        height: var(--ctrl-switch-size-spot);
        width: var(--ctrl-switch-size-spot);
        left: -1px; top: 50%;
        transform: translateY(-50%);
        background: var(--ctrl-switch-off);
        box-shadow: 0 1px 4px 0 var(--ctrl-switch-shadow);
        border-radius: 50%;
      }
      :host input { ${WUPcssHidden} }
      :host input:checked + * + * {
        background-color: var(--ctrl-switch-on-bg);
      }
      :host input:checked + * + *:before {
        background: var(--ctrl-switch-on);
        left: calc(100% - var(--ctrl-switch-size-w) + var(--ctrl-switch-size-spot) + 1px);
      }
      :host[reverse] label {
        flex-direction: row-reverse;
      }
      :host[reverse] label>span {
        margin-left: 0;
        margin-right: 0.5em;
      }
      :host[reverse] label>strong {
        margin-right: auto;
      }
      @media not all and (prefers-reduced-motion) {
        :host label>span { transition: background-color var(--anim); }
        :host label>span:before { transition: left var(--anim); }
      }
      @media (hover: hover) and (pointer: fine) {
        :host:hover label>span:before {
           box-shadow: 0 0 4px 0 var(--ctrl-focus);
        }
      }`;
    }
    static get observedOptions() {
        const arr = super.observedOptions;
        arr.push("reverse");
        return arr;
    }
    static get observedAttributes() {
        const arr = super.observedAttributes;
        arr.push("reverse", "defaultchecked");
        return arr;
    }
    static $isEqual(v1, v2) {
        return !!v1 === !!v2;
    }
    static $defaults = {
        ...WUPBaseControl.$defaults,
        validationRules: { ...WUPBaseControl.$defaults.validationRules },
    };
    get $value() {
        return !!super.$value;
    }
    set $value(v) {
        super.$value = !!v;
    }
    parse(text) {
        return text === "1" || text.toLowerCase() === "true";
    }
    valueToUrl(v) {
        return v ? "1" : null;
    }
    renderControl() {
        this.$refInput.id = this.#ctr.$uniqueId;
        this.$refLabel.setAttribute("for", this.$refInput.id);
        this.$refInput.type = "checkbox";
        this.$refLabel.appendChild(this.$refInput);
        this.$refLabel.appendChild(this.$refTitle);
        this.$refLabel.appendChild(document.createElement("span")); // for icon
        this.appendChild(this.$refLabel);
    }
    gotReady() {
        super.gotReady();
        this.appendEvent(this.$refInput, "input", (e) => this.gotInput(e));
    }
    /** Called when user changes value via click or keyboard */
    gotInput(e) {
        const el = e.target;
        if (this.$isReadOnly) {
            el.checked = !el.checked;
        }
        else {
            this.setValue(el.checked, 3 /* SetValueReasons.userInput */);
        }
    }
    setValue(v, reason) {
        const r = super.setValue(v, reason);
        this.$refInput.checked = !!v;
        return r;
    }
    gotChanges(propsChanged) {
        super.gotChanges(propsChanged);
        this._opts.reverse = this.getAttr("reverse", "bool");
        this.setAttr("reverse", this._opts.reverse, true);
        if (!propsChanged) {
            this.$initValue = this.getAttr("defaultchecked", "bool", this.$initValue);
        }
        else if (propsChanged.includes("defaultchecked")) {
            this.$initValue = this.getAttr("defaultchecked", "bool", undefined);
        }
    }
    gotOptionsChanged(e) {
        this._isStopChanges = true;
        e.props.includes("reverse") && this.setAttr("reverse", this._opts.reverse, true);
        super.gotOptionsChanged(e);
    }
    gotFormChanges(propsChanged) {
        super.gotFormChanges(propsChanged);
        this.setAttr.call(this.$refInput, "aria-readonly", this.$isReadOnly);
    }
}
customElements.define(tagName, WUPSwitchControl);
