import MaskTextInput from "./text.mask";
import { onEvent } from "../indexHelpers";
import { WUPcssIcon } from "../styles";
import WUPBaseControl from "./baseControl";
import TextHistory from "./text.history";
const emailReg = /^(([^<>()[\]\\.,;:\s@"]+(\.[^<>()[\]\\.,;:\s@"]+)*)|(".+"))@((\[[0-9]{1,3}\.[0-9]{1,3}\.[0-9]{1,3}\.[0-9]{1,3}])|(([a-zA-Z\-0-9]+\.)+[a-zA-Z]{2,}))$/;
const tagName = "wup-text";
/** Form-control with text-input
 * @example
  const el = document.createElement("wup-text");
  el.$options.name = "email";
  el.$options.validations = { required: true, max: 100, email: true };
  const form = document.body.appendChild(document.createElement("wup-form"));
  form.appendChild(el);
  // or HTML
  <wup-form>
    <wup-text name="firstName" initvalue="Donny" validations="myValidations"/>
  </wup-form>;
 * @tutorial innerHTML @example
 * <label>
 *   <span> // extra span requires to use with icons via label:before, label:after without adjustments
 *      <input/>
 *      <strong>{$options.label}</strong>
 *   </span>
 *   <button clear/>
 * </label>
 */
export default class WUPTextControl extends WUPBaseControl {
    /** Returns this.constructor // watch-fix: https://github.com/Microsoft/TypeScript/issues/3841#issuecomment-337560146 */
    #ctr = this.constructor;
    static get observedOptions() {
        const arr = super.observedOptions;
        arr.push("clearButton", "maskholder", "mask", "prefix", "postfix");
        return arr;
    }
    static get observedAttributes() {
        const arr = super.observedAttributes;
        arr.push("maskholder", "mask", "prefix", "postfix");
        return arr;
    }
    static get $styleRoot() {
        return `:root {
      --ctrl-clear-hover: rgba(255,0,0,0.1);
      --ctrl-clear-hover-size: 22px;
     }`;
    }
    static get $style() {
        return `${super.$style}
        :host {
          contain: style;
          cursor: text;
        }
        :host label > span {
          width: 100%;
          position: relative;${ /* to position <strong /> relative to input */""}
          display: flex;
          flex-direction: row-reverse;${ /* required for reading <strong /> 1st */""}
        }
        :host input,
        :host textarea,
        :host [maskholder],
        :host [prefix],
        :host [postfix] {
          padding: var(--ctrl-padding);
          padding-left: 0;
          padding-right: 0;
          font: inherit;
          color: inherit;
          margin: 0;
          text-overflow: ellipsis;
          overflow: hidden;
          white-space: nowrap;
        }
        :host input,
        :host textarea,
        :host [contenteditable=true],
        :host [maskholder],
        :host [postfix] {
          width: 100%;
          box-sizing: border-box;
          border: none;
          background: none;
          outline: none;
        }
        :host [prefix],
        :host [postfix] {
          color: var(--ctrl-label);
          flex-shrink: 0;
        }
        :host [maskholder],
        :host [prefix],
        :host [postfix] {
          display: none;
          pointer-events: none;
          text-overflow: initial;
          white-space: pre;
        }
        :host [postfix] {
          position: absolute;
        }
        :host [maskholder] {
          position: absolute;
          opacity: 0.65;
        }
        :host [maskholder]>i,
        :host [postfix]>i {
          visibility: hidden;
          font: inherit;
          white-space: pre;
        }
        :host input:-webkit-autofill,
        :host textarea:-webkit-autofill,
        :host [contenteditable=true]:-webkit-autofill {
          font: inherit;
          -webkit-background-clip: text;
        }
        :host input:autofill,
        :host textarea:autofill,
        :host [contenteditable=true]:autofill {
          font: inherit;
          background-clip: text;
        }
        :host strong {
          display: block;
          position: absolute;
          top: 50%;
          left: 0;
          padding: 0;
          margin: 0;
          box-sizing: border-box;
          max-width: 100%;
          text-overflow: ellipsis;
          overflow: hidden;
          white-space: nowrap;
          transform-origin: top left;
          transform: translateY(-50%);
          font-weight: normal;
          text-decoration: none;
        }
        @media not all and (prefers-reduced-motion) {
          :host strong {
            transition: top var(--anim), transform var(--anim), color var(--anim);
          }
        }
        :host input:not(:focus)::placeholder,
        :host textarea:not(:focus)::placeholder {
          color: transparent;
        }
        :host:focus-within strong,
        :host input:not(:placeholder-shown) + strong,
        :host textarea:not(:placeholder-shown) + strong,
        :host [contenteditable=true]:not(:empty) + strong,
        :host legend {
          top: 0.2em;
          transform: scale(0.9);
        }
        :host:focus-within [maskholder],
        :host:focus-within [prefix],
        :host:focus-within [postfix],
        :host input:not(:placeholder-shown) ~ [prefix],
        :host input:not(:placeholder-shown) ~ [postfix],
        :host textarea:not(:placeholder-shown) ~ [prefix],
        :host textarea:not(:placeholder-shown) ~ [postfix] {
          display: inline-block;
        }
        /* style for icons */
        :host label button,
        :host label button:after,
        :host label:after,
        :host label:before {
          ${WUPcssIcon}
          -webkit-mask-image: none;
          mask-image: none;
        }
        :host label:after {
          cursor: pointer;
          margin-right: calc(var(--ctrl-icon-size) / -2);
        }
        :host label:before {
          cursor: pointer;
          margin-left: calc(var(--ctrl-icon-size) / -2);
        }
        :host label button {
          contain: strict;
          position: relative;
          z-index: 1;
          font-size: inherit;
        }
        :host label>span + button {
          margin-right: -0.5em;
        }
        :host button[clear] {
          align-self: center;
          background: none;
          cursor: pointer;
          --ctrl-icon-img: var(--wup-icon-cross);
        }
        :host button[clear]:after {
          content: "";
          padding: 0;
          -webkit-mask-image: var(--ctrl-icon-img);
          mask-image: var(--ctrl-icon-img);
        }
        :host button[clear=back]:after {
          --ctrl-icon-img: var(--wup-icon-back);
        }
        :host button[clear]:after,
        :host button[clear]:before {
          position: absolute;
          top: 50%; left: 50%;
          transform: translate(-50%, -50%);
          width: 100%;
        }
        :host button[clear]:before {
          width: var(--ctrl-clear-hover-size);
          padding-top: var(--ctrl-clear-hover-size);
        }
        :host[disabled] button[clear],
        :host[readonly] button[clear],
        :host[required] button[clear] {
          display: none;
        }
        @media (hover: hover) and (pointer: fine) {
          :host button[clear]:hover {
            box-shadow: none;
          }
          :host button[clear]:hover:before {
            content: "";
            border-radius: 50%;
            box-shadow: inset 0 0 0 99999px var(--ctrl-clear-hover);
          }
          :host button[clear]:hover:after {
            background-color: var(--ctrl-err-text);
          }
          :host[readonly] button[clear] {
            pointer-events: none;
          }
        }`;
    }
    static $defaults = {
        ...WUPBaseControl.$defaults,
        selectOnFocus: false,
        clearButton: true,
        validationRules: {
            ...WUPBaseControl.$defaults.validationRules,
            min: (v, setV) => (v === undefined || v.length < setV) && `Min length is ${setV} characters`,
            max: (v, setV) => (v === undefined || v.length > setV) && `Max length is ${setV} characters`,
            email: (v, setV) => setV && (!v || !emailReg.test(v)) && "Invalid email address",
            _mask: (_v, setV, c) => {
                const { refMask } = c;
                // WARN: mask ignores value === undefined otherwise it doesn't work with controls that $value !== input.value
                return !!refMask && refMask.value !== refMask.prefix && !refMask.isCompleted && (setV || "Incomplete value");
            },
            _parse: (_v, setV) => setV || "Invalid value",
        },
    };
    $refBtnClear;
    $refMaskholder;
    $refPrefix;
    $refPostfix;
    constructor() {
        super();
        this.$refInput.placeholder = " ";
    }
    // eslint-disable-next-line @typescript-eslint/no-unused-vars
    parse(text) {
        return (text || undefined);
    }
    /** Called before parseInput on gotInput event */
    // eslint-disable-next-line @typescript-eslint/no-unused-vars
    canParseInput(_text) {
        return true;
    }
    /** Called to parse input text to value (related to locale or pointed format) */
    parseInput(text) {
        return this.parse(text.trim());
    }
    /** Returns true if need to use custom undo/redo (required when input somehow formatted/masked) */
    canHandleUndo() {
        return true;
    }
    get validations() {
        const vls = super.validations || {};
        if (this._opts.mask && vls._mask === undefined)
            vls._mask = ""; // enable validation mask based on option mask
        if (this._onceErrName === "_parse")
            vls._parse = "";
        return vls;
    }
    renderControl() {
        this.$refInput.type = "text";
        this.$refInput.id = this.#ctr.$uniqueId;
        this.$refLabel.setAttribute("for", this.$refInput.id);
        const s = this.$refLabel.appendChild(document.createElement("span"));
        s.appendChild(this.$refInput); // input appended to span to allow user user :after,:before without padding adjust
        s.appendChild(this.$refTitle);
        this.appendChild(this.$refLabel);
    }
    /** Create & append element to control */
    renderBtnClear() {
        const bc = document.createElement("button");
        const span = this.$refLabel.querySelector("span");
        if (span.nextElementSibling) {
            this.$refLabel.insertBefore(bc, span.nextElementSibling);
        }
        else {
            this.$refLabel.appendChild(bc);
        }
        bc.setAttribute("clear", "");
        bc.setAttribute("tabindex", -1);
        bc.setAttribute("aria-hidden", true);
        bc.setAttribute("type", "button");
        onEvent(bc, "click", (e) => {
            e.preventDefault(); // prevent from submit
            this.clearValue();
        }, { passive: false });
        return bc;
    }
    /** Add/update/remove prefix part */
    renderPrefix(text) {
        let el = this.$refPrefix;
        if (!text) {
            if (el) {
                el.remove();
                delete this.$refPrefix;
            }
        }
        else {
            if (!el) {
                el = document.createElement("span");
                el.setAttribute("prefix", "");
                this.$refLabel.firstElementChild.append(el);
                this.$refPrefix = el;
            }
            el.textContent = text;
        }
        // because postfix depends on prefix
        this.renderPostfix(this._opts.postfix);
    }
    /** Add/update or remove prefix part */
    renderPostfix(text) {
        let el = this.$refPostfix;
        if (!text) {
            if (el) {
                el.remove();
                delete this.$refPostfix;
            }
        }
        else {
            if (!el) {
                el = document.createElement("span");
                el.setAttribute("postfix", "");
                el.appendChild(document.createElement("i"));
                el.append("");
                this.$refLabel.firstElementChild.append(el);
                this.$refPostfix = el;
            }
            el.firstChild.textContent =
                (this.$isFocused && this.$refMaskholder?.textContent) ||
                    (this.$refPrefix?.textContent || "") + this.$refInput.value;
            el.lastChild.textContent = text;
        }
    }
    /** Custom history undo/redo */
    _refHistory;
    gotFocus(ev) {
        const arr = super.gotFocus(ev);
        if (this.canHandleUndo()) {
            this._refHistory = this._refHistory ?? new TextHistory(this.$refInput);
        }
        const r = this.appendEvent(this.$refInput, "input", (e) => {
            // (e as WUP.Text.GotInputEvent).setValuePrevented = false;
            // (e as WUP.Text.GotInputEvent).preventSetValue = () => ((e as WUP.Text.GotInputEvent).setValuePrevented = true);
            this.gotInput(e);
        });
        const r2 = this.appendEvent(this.$refInput, "beforeinput", (e) => this.gotBeforeInput(e), { passive: false });
        if (!this.$refInput.readOnly) {
            let canSelectAll = this._opts.selectOnFocus;
            if (this._opts.mask) {
                this.maskInputProcess(null); // to apply prefix + maskholder
                canSelectAll = canSelectAll && this.refMask.isCompleted;
                this.renderPostfix(this._opts.postfix);
                if (!canSelectAll) {
                    const end = this.$refInput.value.length;
                    this.$refInput.setSelectionRange(end, end); // move cursor to the end
                }
            }
            canSelectAll && this.$refInput.select();
        }
        const hasOnlyNums = this.refMask?.chunks.every((c) => !c.isVar || c.pattern[0] !== "*");
        this.setAttr.call(this.$refInput, "inputmode", hasOnlyNums ? "numeric" : "");
        arr.push(() => setTimeout(r)); // timeout required to handle Event on gotFocusLost
        arr.push(r2);
        return arr;
    }
    gotFocusLost() {
        this.#declineInputEnd?.call(this);
        if (this.refMask) {
            if (this.refMask.prefix && this.refMask.value === this.refMask.prefix) {
                this.$refInput.value = ""; // rollback prefix/postfix if user types nothing
                delete this.$refInput._maskPrev;
            }
            this.renderPostfix(this._opts.postfix); // postfix depends on maskholder
        }
        super.gotFocusLost();
    }
    gotChanges(propsChanged) {
        // apply mask options
        this._opts.mask = this.getAttr("mask");
        this._opts.maskholder = this.getAttr("maskholder");
        if (!this._opts.mask || this._opts.mask !== this.refMask?.pattern) {
            delete this.refMask; // delete if mask is removed or changed (it's recovered again on event)
        }
        if (this._opts.mask && this._opts.maskholder == null) {
            this.refMask = this.refMask ?? new MaskTextInput(this._opts.mask, "");
            this._opts.maskholder = this.refMask.chunks.map((c) => c.pattern).join("");
        }
        if ((!this._opts.maskholder || !this._opts.mask) && this.$refMaskholder) {
            this.$refMaskholder.remove();
            delete this.$refMaskholder;
        }
        super.gotChanges(propsChanged);
        if (this._opts.clearButton) {
            this.$refBtnClear = this.$refBtnClear || this.renderBtnClear();
        }
        else if (this.$refBtnClear) {
            this.$refBtnClear.remove();
            this.$refBtnClear = undefined;
        }
        if (!propsChanged || propsChanged.includes("prefix")) {
            this._opts.prefix = this.getAttr("prefix");
            this.renderPrefix(this._opts.prefix);
        }
        if (!propsChanged || propsChanged.includes("postfix")) {
            this._opts.postfix = this.getAttr("postfix");
            this.renderPostfix(this._opts.postfix);
        }
    }
    gotKeyDown(e) {
        super.gotKeyDown(e);
        this._refHistory?.handleKeyDown(e);
    }
    /** Value-text before input changed - used with declineInput */
    _beforeSnap;
    /** Handler of 'beforeinput' event */
    gotBeforeInput(e) {
        const isUndoRedo = this._refHistory?.handleBeforeInput(e);
        this.#declineInputEnd?.call(this);
        this._beforeSnap = TextHistory.historyToSnapshot(this.$refInput.value, this.$refInput.selectionStart || 0);
        setTimeout(() => delete this._beforeSnap);
        if (!isUndoRedo && this._opts.mask) {
            this.refMask = this.refMask ?? new MaskTextInput(this._opts.mask, e.target.value);
            this.refMask.handleBeforeInput(e);
        }
    }
    #inputTimer;
    /** Called when user types text OR when need to apply/reset mask (on focusGot, focusLost) */
    gotInput(e) {
        this._refHistory?.handleInput(e);
        const el = e.target;
        let txt = el.value;
        if (this._opts.mask) {
            const prev = el._maskPrev?.value;
            txt = this.maskInputProcess(e);
            if (txt === prev) {
                // this.renderPostfix(this._opts.postfix);
                return; // skip because no changes from previous action
            }
        }
        // if (e.setValuePrevented) { // use canParse instead
        //   return;
        // }
        const canParse = !txt || this.canParseInput(txt);
        let v = this.$value;
        let errMsg = "";
        if (canParse) {
            try {
                v = !txt ? undefined : this.parseInput(txt);
            }
            catch (err) {
                errMsg = err.message || true;
            }
        }
        this.renderPostfix(this._opts.postfix);
        if (this.#declineInputEnd && (!canParse || errMsg)) {
            return; // don't allow changes if user types wrong char
        }
        const act = () => {
            if (errMsg) {
                this.validateOnce({ _parse: this.validations?._parse || "" }, true);
            }
            else if (canParse) {
                this.setValue(v, 3 /* SetValueReasons.userInput */, true);
            }
            else if (this._opts.mask) {
                this.validateOnce({ _mask: this.validations?._mask || "" });
            }
        };
        this._validTimer && clearTimeout(this._validTimer);
        this.#inputTimer && clearTimeout(this.#inputTimer);
        if (this._opts.debounceMs) {
            this.#inputTimer = setTimeout(act, this._opts.debounceMs);
        }
        else {
            act();
        }
    }
    /** Mask object to proccess mask on input */
    refMask;
    /** Called to apply mask-behavior (on "input" event) */
    maskInputProcess(e) {
        const el = this.$refInput;
        const v = el.value;
        const { mask } = this._opts;
        this.refMask = this.refMask ?? new MaskTextInput(mask, "");
        const mi = this.refMask;
        const isFocused = this.$isFocused;
        if (!v && !isFocused) {
            el.value = v;
            mi.parse(v);
            return v; // ignore mask prefix/postfix if user isn't touched input; it appends only by focusGot
        }
        let declinedAdd = 0;
        let pos = el.selectionStart || 0;
        if (!e) {
            mi.parse(v);
            pos = mi.value.length; // fix case: mask with prefix + call .clearValue()
        }
        else {
            const r = mi.handleInput(e);
            declinedAdd = r.declinedAdd;
            pos = r.position;
        }
        if (declinedAdd) {
            this.declineInput(pos, mi.value); // fix when ###: "12|" + "3b" => 123|
        }
        else if (el.value !== mi.value) {
            el.value = mi.value;
            document.activeElement === el && el.setSelectionRange(pos, pos); // without checking on focus setSelectionRange sets focus on Safari
        }
        isFocused && this.renderMaskHolder(this._opts.maskholder, mi.leftLength - declinedAdd);
        return mi.value;
    }
    /** Add/update maskholder or skip if it's not defined */
    renderMaskHolder(text, leftLength) {
        if (!text) {
            return;
        }
        if (!this.$refMaskholder) {
            const m = document.createElement("span");
            m.setAttribute("aria-hidden", "true");
            m.setAttribute("maskholder", "");
            m.appendChild(document.createElement("i"));
            m.append("");
            this.$refInput.parentElement.prepend(m);
            this.$refMaskholder = m;
        }
        this.$refMaskholder.firstChild.textContent = (this._opts.prefix || "") + this.$refInput.value;
        this.$refMaskholder.lastChild.textContent = text.substring(text.length - leftLength);
    }
    #declineInputEnd;
    /** Make undo for input after 100ms when user types not allowed chars
     * point nextCaret for custom undo value
     * @tutorial Troubleshooting
     * * declineInput doesn't trigger beforeinput & input events (do it manually if required) */
    declineInput(nextCaret, nextValue) {
        if (!this._beforeSnap) {
            return;
        }
        const { v, pos } = TextHistory.historyFromSnapshot(this._beforeSnap);
        nextCaret ??= pos;
        nextValue ??= v;
        delete this._beforeSnap;
        const isChanged = v !== nextValue;
        isChanged && nextValue != null && this._refHistory?.save(v, nextValue); // update history if last value is different
        this.#declineInputEnd = () => {
            this.#declineInputEnd = undefined;
            clearTimeout(t);
            this.$refInput.value = nextValue;
            this.$refInput.setSelectionRange(nextCaret, nextCaret);
            this.refMask && this.renderMaskHolder(this._opts.maskholder, this.refMask.leftLength);
            this.renderPostfix(this._opts.postfix);
            v === nextValue && this._refHistory?.removeLast();
        };
        const t = setTimeout(this.#declineInputEnd, 100);
    }
    setValue(v, reason, skipInput = false) {
        !skipInput && this.setInputValue(v, reason);
        const isChanged = super.setValue(v, reason);
        this._isValid !== false && this.goHideError();
        return isChanged;
    }
    /** Called to update/reset value for <input/> */
    // eslint-disable-next-line @typescript-eslint/no-unused-vars
    setInputValue(v, reason) {
        const prev = this.$refInput.value;
        const str = v != null ? v.toString() : "";
        this.$refInput.value = str;
        this._opts.mask && this.maskInputProcess(null);
        str !== prev && this._refHistory?.save(prev, this.$refInput.value);
        this.renderPostfix(this._opts.postfix);
        this._onceErrName === this._errName && this.goHideError(); // hide mask-message because value has higher priority than inputValue
    }
    setClearState() {
        const next = super.setClearState();
        if (this.$refBtnClear) {
            this.$refBtnClear.setAttribute("clear", this.#ctr.$isEmpty(next) ? "" : "back");
        }
        return next;
    }
    _onceErrName;
    /** Called when inputValue != $value and need to show error on the fly by input-change */
    validateOnce(rule, force = false) {
        const prev = this._wasValidNotEmpty;
        if (force) {
            this._wasValidNotEmpty = true; // to ignore onChangeSmart and show error anyway
        }
        // redefine prototype getter once & fire validation
        Object.defineProperty(this, "validations", { configurable: true, value: rule });
        this.validateAfterChange();
        delete this.validations; // rollback to previous
        if (force) {
            this._wasValidNotEmpty = prev; // rollback to previous
        }
        this._onceErrName = this._errName;
    }
    goHideError() {
        this._onceErrName = undefined;
        super.goHideError();
    }
}
customElements.define(tagName, WUPTextControl);
// todo example how to create bult-in dropdown before the main input (like phone-number with ability to select countryCode)
// gotInput > setMask > parseValue >... setValue ....> toString > setInput > setMask
