import WUPBaseElement from "./baseElement";
import WUPPopupElement from "./popup/popupElement";
import { Animations, HideCases, ShowCases } from "./popup/popupElement.types";
declare const tagName = "wup-dropdown";
declare global {
    namespace WUP.Dropdown {
        interface Defaults extends WUP.Popup.Defaults, Pick<WUP.Popup.Options, "minWidthByTarget" | "minHeightByTarget"> {
            /** Animation that applied to popup;
             * @defaultValue `Animations.drawer`
             * @tutorial Troubleshooting
             * * to change option for specific element change it for `<wup-popup/>` direclty after timeout
             * @example setTimeout(() => this.$refPopup.$options.animation = Animations.stack) */
            animation: Animations;
            /** Case when popup need to show;
             * @defaultValue `ShowCases.onClick | ShowCases.onFocus`
             * @tutorial Troubleshooting
             * * to change option for specific element change it for `<wup-popup/>` directly after timeout
             * @example setTimeout(() => this.$refPopup.$options.showCase = ShowCases.onFocus | ShowCases.onClick) */
            showCase: ShowCases;
            /** Hide on popup click
             * @defaultValue true */
            hideOnPopupClick: boolean;
            /** Placement rules relative to target;
             * @defaultValue `[
                  WUPPopupElement.$placements.$bottom.$start,
                  WUPPopupElement.$placements.$bottom.$end,
                  WUPPopupElement.$placements.$top.$start,
                  WUPPopupElement.$placements.$top.$end,
                  WUPPopupElement.$placements.$bottom.$start.$resizeHeight,
                  WUPPopupElement.$placements.$bottom.$end.$resizeHeight,
                  WUPPopupElement.$placements.$top.$start.$resizeHeight,
                  WUPPopupElement.$placements.$top.$end.$resizeHeight ]` */
            placement: Array<WUP.Popup.Place.PlaceFunc>;
            /** Sets minWidth 100% of targetWidth; it can't be more than css-style min-width
             * @defaultValue true */
            minWidthByTarget: boolean;
            /** Sets minHeight 100% of targetWidth; it can't be more than css-style min-height
             *  @defaultValue true */
            minHeightByTarget: boolean;
        }
        interface Options extends Defaults {
        }
        interface Attributes {
        }
        interface JSXProps<C = WUPDropdownElement> extends WUP.Base.JSXProps<C>, Attributes {
        }
    }
    interface HTMLElementTagNameMap {
        [tagName]: WUPDropdownElement;
    }
    namespace JSX {
        interface IntrinsicElements {
            /** Dropdown element
             *  @see {@link WUPDropdownElement} */
            [tagName]: WUP.Dropdown.JSXProps;
        }
    }
}
/** Dropdown element
 * @example
 * ```html
  <wup-dropdown class="custom">
    <button type="button">Click Me</button>
    <wup-popup placement="bottom-middle" animation="stack">
      <ul>
        <li><button type="button">A</button></li>
        <li><button type="button">B</button></li>
        <li><button type="button">C</button></li>
      </ul>
    </wup-popup>
  </wup-dropdown>
 * ```
 * @tutorial Troubleshooting
 * * in comparison with Combobox Dropdown doesn't contain input but can contain in popup any items (like nav-menu etc.) */
export default class WUPDropdownElement<TOptions extends WUP.Dropdown.Options = WUP.Dropdown.Options> extends WUPBaseElement<TOptions> {
    #private;
    static get $style(): string;
    /** Default options applied to every element. Change it to configure default behavior
     * * @tutorial Troubleshooting
     * * Popup-related options are not observed so to change it use `WUPDropdownElement.$defaults` or `element.$refPopup.$options` direclty */
    static $defaults: WUP.Dropdown.Defaults;
    /** Reference to nested HTMLElement tied with $options.label */
    $refTitle: HTMLElement;
    /** Reference to popupMenu */
    $refPopup: WUPPopupElement<WUP.Popup.Options, WUP.Popup.EventMap>;
    /** Reference to list */
    $refMenu: HTMLElement;
    protected gotReady(): void;
    /** Custom function to override default `WUPPopupElement.prototype.goHide` */
    protected goShowPopup(showCase: ShowCases, ev: MouseEvent | FocusEvent | null): boolean | Promise<boolean>;
    /** Custom function to override default `WUPPopupElement.prototype.goHide` */
    protected goHidePopup(hideCase: HideCases, ev: MouseEvent | FocusEvent | KeyboardEvent | null): boolean | Promise<boolean>;
}
export {};
