import { HideCases, ShowCases } from "./popupElement.types";
export interface PopupListenerOptions {
    target: HTMLElement | SVGElement;
    showCase?: ShowCases;
    hoverShowTimeout?: number;
    hoverHideTimeout?: number;
    focusDebounceMs?: number;
}
/** Listen for target according to showCase. If target removed call stopListen + don't forget to remove popup yourself;
 * @tutorial Troubleshooting/rules:
 * * To allow user hide/show popup before previous action is ended provide for onShow and onHide-
 *   callbacks results as fast as it's possible without waiting for animation */
export default class PopupListener {
    #private;
    options: PopupListenerOptions;
    onShow: (showCase: ShowCases, ev: MouseEvent | FocusEvent | null) => HTMLElement | null | Promise<HTMLElement | null>;
    onHide: (hideCase: HideCases, ev: MouseEvent | FocusEvent | KeyboardEvent | null) => boolean | Promise<boolean>;
    static $defaults: {
        showCase: ShowCases;
        hoverShowTimeout: number;
        hoverHideTimeout: number;
    };
    constructor(options: PopupListenerOptions, onShow: (showCase: ShowCases, ev: MouseEvent | FocusEvent | null) => HTMLElement | null | Promise<HTMLElement | null>, onHide: (hideCase: HideCases, ev: MouseEvent | FocusEvent | KeyboardEvent | null) => boolean | Promise<boolean>);
    openedEl: Element | null;
    isShowing?: boolean | Promise<unknown>;
    isHiding?: boolean | Promise<unknown>;
    /** Called on show */
    show(showCase: ShowCases, e: MouseEvent | FocusEvent | null): Promise<void>;
    /** Called on hide */
    hide(hideCase: HideCases, e: MouseEvent | FocusEvent | KeyboardEvent | null): Promise<void>;
    /** Returns focus back during the hiding */
    focusBack(relatedTarget: Element | EventTarget | {
        focus: () => void;
    } | null): void;
    /** Returns whether el is part of another  */
    includes(this: Element, el: unknown): boolean;
    /** Returns whether element if part of openedElement */
    isMe(el: unknown, openedEl?: Element | null): boolean;
    /** Single event handler */
    handleEventsDocument(e: Event): void;
    /** Single event handler */
    handleEvents(ev: Event): void;
    /** Called on init to apply eventListeners */
    listen(): void;
    /** Called to add extra events when show */
    listenOnShow(): void;
    /** Called to remove extra events when hide */
    disposeOnHide(): void;
    /** Remove all event listeners */
    stopListen(): void;
}
